@@ -7,16 +7,16 @@ This module provides Basic and Digest HTTP authentication for Flask routes.
 :copyright: (C) 2014 by Miguel Grinberg.
 :license:   MIT, see LICENSE for more details.
 """
-
+import hmac
 from base64 import b64decode
 from functools import wraps
 from hashlib import md5
 from random import Random, SystemRandom
 from flask import request, make_response, session, g, Response
 from werkzeug.datastructures import Authorization
-from werkzeug.security import safe_str_cmp
 
-__version__ = '4.2.0'
+
+__version__ = '4.4.0'
 
 
 class HTTPAuth(object):
@@ -37,6 +37,18 @@ class HTTPAuth(object):
         self.get_password(default_get_password)
         self.error_handler(default_auth_error)
 
+    def is_compatible_auth(self, headers):
+        if self.header is None or self.header == 'Authorization':
+            try:
+                scheme, _ = request.headers.get('Authorization', '').split(
+                    None, 1)
+            except ValueError:
+                # malformed Authorization header
+                return False
+            return scheme == self.scheme
+        else:
+            return self.header in headers
+
     def get_password(self, f):
         self.get_password_callback = f
         return f
@@ -169,9 +181,10 @@ class HTTPAuth(object):
         return login_required_internal
 
     def username(self):
-        if not request.authorization:
+        auth = self.get_auth()
+        if not auth:
             return ""
-        return request.authorization.username
+        return auth.username
 
     def current_user(self):
         if hasattr(g, 'flask_httpauth_user'):
@@ -205,9 +218,14 @@ class HTTPBasicAuth(HTTPAuth):
             username, password = b64decode(credentials).split(b':', 1)
         except (ValueError, TypeError):
             return None
+        try:
+            username = username.decode('utf-8')
+            password = password.decode('utf-8')
+        except UnicodeDecodeError:
+            username = None
+            password = None
         return Authorization(
-            scheme, {'username': username.decode('utf-8'),
-                     'password': password.decode('utf-8')})
+            scheme, {'username': username, 'password': password})
 
     def authenticate(self, auth, stored_password):
         if auth:
@@ -228,7 +246,7 @@ class HTTPBasicAuth(HTTPAuth):
                                                               client_password)
         return auth.username if client_password is not None and \
             stored_password is not None and \
-            safe_str_cmp(client_password, stored_password) else None
+            hmac.compare_digest(client_password, stored_password) else None
 
 
 class HTTPDigestAuth(HTTPAuth):
@@ -257,7 +275,7 @@ class HTTPDigestAuth(HTTPAuth):
             session_nonce = session.get("auth_nonce")
             if nonce is None or session_nonce is None:
                 return False
-            return safe_str_cmp(nonce, session_nonce)
+            return hmac.compare_digest(nonce, session_nonce)
 
         def default_generate_opaque():
             session["auth_opaque"] = _generate_random()
@@ -267,7 +285,7 @@ class HTTPDigestAuth(HTTPAuth):
             session_opaque = session.get("auth_opaque")
             if opaque is None or session_opaque is None:  # pragma: no cover
                 return False
-            return safe_str_cmp(opaque, session_opaque)
+            return hmac.compare_digest(opaque, session_opaque)
 
         self.generate_nonce(default_generate_nonce)
         self.generate_opaque(default_generate_opaque)
@@ -326,7 +344,7 @@ class HTTPDigestAuth(HTTPAuth):
         ha2 = md5(a2.encode('utf-8')).hexdigest()
         a3 = ha1 + ":" + auth.nonce + ":" + ha2
         response = md5(a3.encode('utf-8')).hexdigest()
-        return safe_str_cmp(response, auth.response)
+        return hmac.compare_digest(response, auth.response)
 
 
 class HTTPTokenAuth(HTTPAuth):
@@ -362,21 +380,12 @@ class MultiAuth(object):
         def login_required_internal(f):
             @wraps(f)
             def decorated(*args, **kwargs):
-                selected_auth = None
-                if 'Authorization' in request.headers:
-                    try:
-                        scheme, creds = request.headers[
-                            'Authorization'].split(None, 1)
-                    except ValueError:
-                        # malformed Authorization header
-                        pass
-                    else:
-                        for auth in self.additional_auth:
-                            if auth.scheme == scheme:
-                                selected_auth = auth
-                                break
-                if selected_auth is None:
-                    selected_auth = self.main_auth
+                selected_auth = self.main_auth
+                if not self.main_auth.is_compatible_auth(request.headers):
+                    for auth in self.additional_auth:
+                        if auth.is_compatible_auth(request.headers):
+                            selected_auth = auth
+                            break
                 return selected_auth.login_required(role=role,
                                                     optional=optional
                                                     )(f)(*args, **kwargs)
